"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const util_1 = require("./util");
const log = (0, debug_1.default)("dylink:function-table");
const GROW_THRESH = 50;
class FunctionTable {
    constructor(table) {
        // indexes into the function table that are null, but
        // maybe less than nextTablePos.
        this.availableTableIndexes = new Set();
        // next available table position at the very end. Use this if
        // availableTableIndexes is empty.
        this.nextTablePos = 0;
        log("constructor");
        this.table = table;
    }
    updateAfterImport() {
        this.nextTablePos = Math.max(0, ...(0, util_1.nonzeroPositions)(this.table)) + 1;
    }
    prepareForImport(tableSize) {
        this.nextTablePos += tableSize;
        if (this.table.length <= this.nextTablePos + GROW_THRESH) {
            this.table.grow(this.nextTablePos + GROW_THRESH - this.table.length);
        }
    }
    getNextTablePos() {
        return this.nextTablePos;
    }
    // gets the next available index, then marks it
    // as no longer available, so it can be used.
    getNextAvailableIndex() {
        for (const index of this.availableTableIndexes) {
            this.availableTableIndexes.delete(index);
            log("getNextAvailableIndex: from availableTableIndexes", index);
            return index;
        }
        const index = this.nextTablePos;
        log("getNextAvailableIndex: from nextTablePos", index);
        this.nextTablePos += 1;
        if (this.table.length <= this.nextTablePos + GROW_THRESH) {
            this.table.grow(this.nextTablePos + GROW_THRESH - this.table.length);
        }
        return index;
    }
    setLater() {
        const index = this.getNextAvailableIndex();
        return {
            index,
            set: (f) => {
                this.set(f, index);
            },
        };
    }
    set(f, _index) {
        const index = _index ?? this.getNextAvailableIndex();
        if (this.table.get(index) != null) {
            throw Error("BUG: trying to set a table index that is already set");
        }
        this.table.set(index, f);
        return index;
    }
    get(index) {
        return this.table.get(index);
    }
    delete(index) {
        this.table.set(index, null);
        this.availableTableIndexes.add(index);
    }
}
exports.default = FunctionTable;
//# sourceMappingURL=function-table.js.map