"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const log = (0, debug_1.default)("stub");
const logUse = (0, debug_1.default)("stub:use"); // log all use of the stub
const logFirst = (0, debug_1.default)("stub:first"); // log first use of the stub
function stubProxy(env, functionViaPointer, type) {
    return new Proxy(env, {
        get(target, key) {
            if (key in target) {
                return Reflect.get(target, key);
            }
            const f = functionViaPointer(key);
            if (f != null) {
                log("using function via pointer for ", key);
                return f;
            }
            if (type == "warn") {
                console.warn(`\n* WARNING: creating UNSAFE stub for ${String(key)}.  Please fix ASAP!`);
            }
            if (logUse.enabled || logFirst.enabled) {
                return (...args) => {
                    logStubUse(key, args);
                    return 0;
                };
            }
            else {
                // faster to not trace or even check, obviously.
                return () => 0;
            }
        },
    });
}
exports.default = stubProxy;
const stubUsed = new Set([]);
function logStubUse(functionName, args) {
    logUse("WARNING: using stub", functionName, args);
    if (logFirst.enabled) {
        if (stubUsed.has(functionName))
            return;
        stubUsed.add(functionName);
    }
    logFirst("WARNING: first use of stub", functionName, args);
}
//# sourceMappingURL=stub.js.map